/*  ---------------------------------------------------------------  */
/*      The HMM-Based Speech Synthesis System (HTS): version 1.1     */
/*                        HTS Working Group                          */
/*                                                                   */
/*                   Department of Computer Science                  */
/*                   Nagoya Institute of Technology                  */
/*                                and                                */
/*    Interdisciplinary Graduate School of Science and Engineering   */
/*                   Tokyo Institute of Technology                   */
/*                      Copyright (c) 2001-2003                      */
/*                        All Rights Reserved.                       */
/*                                                                   */
/*  Permission is hereby granted, free of charge, to use and         */
/*  distribute this software and its documentation without           */
/*  restriction, including without limitation the rights to use,     */
/*  copy, modify, merge, publish, distribute, sublicense, and/or     */
/*  sell copies of this work, and to permit persons to whom this     */
/*  work is furnished to do so, subject to the following conditions: */
/*                                                                   */
/*    1. The code must retain the above copyright notice, this list  */
/*       of conditions and the following disclaimer.                 */
/*                                                                   */
/*    2. Any modifications must be clearly marked as such.           */
/*                                                                   */    
/*  NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSITITUTE OF TECHNOLOGY,  */
/*  HTS WORKING GROUP, AND THE CONTRIBUTORS TO THIS WORK DISCLAIM    */
/*  ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL       */
/*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL NAGOYA INSTITUTE OF TECHNOLOGY, TOKYO INSITITUTE OF        */
/*  TECHNOLOGY, HTS WORKING GROUP, NOR THE CONTRIBUTORS BE LIABLE    */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY        */
/*  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,  */
/*  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTUOUS   */
/*  ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR          */
/*  PERFORMANCE OF THIS SOFTWARE.                                    */
/*                                                                   */
/*  ---------------------------------------------------------------  */
/*    hts_engine.c : a compact HMM-based speech synthesis engine     */
/*                                                                   */ 
/*                                    2003/05/09 by Heiga Zen        */
/*  ---------------------------------------------------------------  */

/*  Standard C Libraries  */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "misc.h"
#include "tree.h"
#include "model.h"
#include "global.h"
#include "vocoder.h"
#include "mlpg.h"
#include "defaults.h"

void usage (char *cmnd, int status)
{
   fprintf (stderr, "\n");
   fprintf (stderr, " %s - A compact HMM-based speech synthesis engine\n",cmnd);
   fprintf (stderr, "\n");
   fprintf (stderr, "  usage:\n");
   fprintf (stderr, "       %s [ options ] [ infile ] > stdout\n", cmnd);
   fprintf (stderr, "  options:\n");
   fprintf (stderr, "       -td tree  : decision trees file for state duration           [N/A]\n");
   fprintf (stderr, "       -tf tree  : decision trees file for Log F0                   [N/A]\n");
   fprintf (stderr, "       -tm tree  : decision trees file for Mel-Cepstrum             [N/A]\n");
   fprintf (stderr, "       -md pdf   : model file for state duration                    [N/A]\n");
   fprintf (stderr, "       -mf pdf   : model file for Log F0                            [N/A]\n");
   fprintf (stderr, "       -mm pdf   : model file for Mel-Cepstrum                      [N/A]\n");
   fprintf (stderr, "       -df win   : window file for culcuration delta of log F0      [N/A]\n");
   fprintf (stderr, "       -dm win   : filename of delta coeffs for Mel-Cepstrum        [N/A]\n");
   fprintf (stderr, "       -of s     : filename of output F0                            [N/A]\n");
   fprintf (stderr, "       -om s     : filename of output mcep                          [N/A]\n");
   fprintf (stderr, "       -or s     : filename of output raw audio (generated speech)  [N/A]\n");
   fprintf (stderr, "       -vs       : use state alignment for duration                 [FALSE]\n");
   fprintf (stderr, "       -vp       : use phoneme alignment for duration               [FALSE]\n");
   fprintf (stderr, "       -a  f     : all-pass constant                                [0.42]\n");
   fprintf (stderr, "       -r  f     : control duration parameter                       [0.0]\n");
   fprintf (stderr, "       -fs f     : multilply f0                                     [1.0]\n");
   fprintf (stderr, "       -fm f     : add f0                                           [0.0]\n");
   fprintf (stderr, "       -u  f     : voiced/unvoiced threshold                        [0.5]\n");
   fprintf (stderr, "       -l  i     : length of generated speech (in second)           [N/A]\n");
   fprintf (stderr, "  infile:\n");
   fprintf (stderr, "       label file\n");
   fprintf (stderr, "  stdout:\n");
   fprintf (stderr, "       label with frame number\n");
   fprintf (stderr, "  note:\n");
   fprintf (stderr, "       option '-d' may be repeated to use multiple\n");
   fprintf (stderr, "       delta parameters\n");
   fprintf (stderr, "       if 'd' is an integer, delta coefficients are\n");
   fprintf (stderr, "       calculated based on a regression formula.\n");
   fprintf (stderr, "\n");
 
   exit (status);
}

void Process ( FILE *, FILE *, FILE *, FILE *, PStream *, PStream *, 
                globalP *, ModelSet *, TreeSet *, VocoderSetup *);

int main(int argc, char **argv)
{
   FILE *labfp=stdin;
   FILE *lf0fp=NULL, *mcepfp=NULL, *rawfp=NULL;
   int i;
   float a,m,s,r,l;
   char *cmnd;
   
   ModelSet     ms;
   TreeSet      ts;
   PStream      mceppst, lf0pst;
   globalP      gp;
   VocoderSetup vs;
   
   /* default value for control parameter */
   gp.RHO      = 0.0; 
   gp.ALPHA    = 0.42;
   gp.F0_STD   = 1.0;
   gp.F0_MEAN  = 0.0;
   gp.UV       = 0.5;
   gp.LENGTH   = 0;
   gp.algnst   = FALSE;
   gp.algnph   = FALSE;
   gp.XIMERA   = FALSE;
   
   /* initialise TreeSet and ModelSet */
   InitTreeSet (&ts);
   InitModelSet(&ms);
   
   /* delta window handler for log f0 */
   lf0pst.dw.fn = (char **) calloc (argc, sizeof (char *));
   lf0pst.dw.num = 1;
   
   /* delta window handler for mel-cepstrum */
   mceppst.dw.fn = (char **) calloc (argc, sizeof (char *));
   mceppst.dw.num = 1;
  
   if ((cmnd = strrchr (argv[0], '/')) == NULL) {
      cmnd = argv[0];
   }
   else {
      cmnd++;
   }
   
   if (argc==1)
      usage(cmnd, 1);
   
   /* command line parsing */
   while (--argc)
      if (**++argv == '-') {
         switch (*(*argv+1)) {
            case 'v': 
               switch (*(*argv+2)) {
                  case 's' : gp.algnst = TRUE;  break;
                  case 'p' : gp.algnph = TRUE;  break;
                  default : usage(cmnd, 1);
               }
               break;
            case 'x': gp.XIMERA = TRUE; break;
            case 't':
               switch (*(*argv+2)) {
                  case 'd' : ts.fp[DUR] = getfp (*(argv+1), "r");  break;
                  case 'f' : 
                  case 'p' : ts.fp[LF0] = getfp (*(argv+1), "r");  break;
                  case 'm' : ts.fp[MCP] = getfp (*(argv+1), "r");  break;
                  default : usage(cmnd, 1); 
               }
               ++argv; --argc;
               break;
            case 'm':
               switch (*(*argv+2)) {
                  case 'd' : ms.fp[DUR] = getfp (*(argv+1), "r");  break;
                  case 'f' : 
                  case 'p' : ms.fp[LF0] = getfp (*(argv+1), "r");  break;
                  case 'm' : ms.fp[MCP] = getfp (*(argv+1), "r");  break;
                  default : usage(cmnd, 1); 
               }
               ++argv; --argc;
               break;
            case 'd':
               switch (*(*argv+2)) {
                  case 'm' :  mceppst.dw.fn[mceppst.dw.num] = *(argv+1);
                              mceppst.dw.num++;
                              break;
                  case 'f' :
                  case 'p' :  lf0pst.dw.fn[lf0pst.dw.num] = *(argv+1);
                              lf0pst.dw.num++;
                              break;
                  default : usage(cmnd, 1); 
               }
               ++argv; --argc;
               break;
            case 'o':
               switch (*(*argv+2)) {
                  case 'r' : rawfp  = getfp (*(argv+1), "w");  break;
                  case 'f' : 
                  case 'p' : lf0fp  = getfp (*(argv+1), "w");  break;
                  case 'm' : mcepfp = getfp (*(argv+1), "w");  break;
                  default : usage(cmnd, 1); 
               }
               ++argv; --argc;
               break;
            case 'h': usage (cmnd, 0); break;
            case 'a':
               a = atof(*++argv);
               if (a<1 && a>0) gp.ALPHA = a; 
               --argc;
               break;
            case 'r':
               r = atof(*++argv);
               if (a<1 && a>-1) gp.RHO = r;
               --argc;
               break;
            case 'f':
               switch (*(*argv+2)) {
                  case 's' :  s = atof(*++argv);  if (s>0) gp.F0_STD=s;     break;
                  case 'm' :  m = atof(*++argv);  if (m>0) gp.F0_MEAN = m;  break;
                  default : usage(cmnd, 1); 
               }
               --argc;
               break;
            case 'u':
               r = atof(*++argv);
               if (a<1 && a>0) gp.UV = r;
               --argc;
               break;
            case 'l':
               a = atof(*++argv);
               if (a<30 && a>0) gp.LENGTH = (int)(a*RATE/FPERIOD);
               --argc;
               break;
            default:
               fprintf (stderr, "%s : Invalid option '%c' !\n", 
                        cmnd, *(*argv + 1));
               usage(cmnd, 1);
         }
      }
      else 
         labfp = getfp(*argv, "r");

   if (ts.fp[DUR] == NULL) {
      fprintf (stderr, "%s : duration treefile is needed!\n", cmnd);
      usage(cmnd, 1);
   }
   
   if (ts.fp[LF0] == NULL) {
      fprintf (stderr, "%s : log F0 treefile is needed!\n", cmnd);
      usage(cmnd, 1);
   }
   
   if (ts.fp[MCP] == NULL) {
      fprintf (stderr, "%s : mcep treefile is needed!\n", cmnd);
      usage(cmnd, 1);
   }
   
   if (ms.fp[DUR] == NULL) {
      fprintf (stderr, "%s : duration modelfile is needed!\n", cmnd);
      usage(cmnd, 1);
   }
   
   if (ms.fp[MCP] == NULL) {
      fprintf (stderr, "%s : mcep modelfile is needed!\n", cmnd);
      usage(cmnd, 1);
   }
   
   if (ms.fp[LF0] == NULL) {
      fprintf (stderr, "%s : log F0 modelfile is needed!\n", cmnd);
      usage(cmnd, 1);
   } 
   
   if (gp.algnst && gp.algnph) {
      fprintf (stderr, "%s : cannot specify both '-vs' and '-vp' options !\n", cmnd);
      usage(cmnd, 1);
   } 
   
   if (gp.LENGTH>0 && gp.RHO!=0.0) {
      fprintf (stderr, "%s : cannot specify both '-r' and '-l' options !\n", cmnd);
      usage(cmnd, 1);
   } 
   
   /* load decision trees for duration, log f0 and mel-cepstrum */
   LoadTreesFile(&ts, DUR);
   LoadTreesFile(&ts, LF0);
   LoadTreesFile(&ts, MCP);
   
   /* load model files for duration, log f0 and mel-cepstrum */
   LoadModelFiles(&ms);
   
   /* if the name of output speech file is not specified, waveform generation won't be generated */
   if (rawfp!=NULL)
      init_vocoder(ms.mcepvsize-1, &vs);

   Process(labfp, rawfp, lf0fp, mcepfp, &mceppst, &lf0pst, &gp, &ms, &ts, &vs);

   return 0;
}

/* OutLabel : output label with frame number or time */
void OutLabel (UttModel *um, Boolean XIMERA) 
{
   Model *m;
   char *tmp;
   int nframe = 0;
   
   for (m=um->mhead; m!=um->mtail; m=m->next) {
      if (XIMERA) {      /* in XIMERA format */
         tmp = strdup(m->name);
         tmp = strchr(tmp,'-')+1;
         *(strchr(tmp,'+')) = '\0';
         fprintf(stdout,"%1.3f  %s\n", (((float)nframe)*FPERIOD)/((float)RATE), tmp);   
      } 
      else               /* in HTK & HTS format */
         fprintf(stdout, "%d %d %s\n", nframe, nframe+m->totaldur,m->name);
      
      nframe += m->totaldur;
   }
}

void Process ( FILE *labfp, FILE *rawfp, FILE *lf0fp, FILE *mcepfp, 
               PStream *mceppst, PStream *lf0pst, globalP *gp, ModelSet *ms, TreeSet *ts, VocoderSetup *vs) 
{
   char buf[1024];
   Tree *tree;
   int state, diffdur=0;
   int start, end;
   int rate, nf;
   float f, mean, var;
   Boolean hastime;
   Model *m;
   UttModel um;
   
   rate = FPERIOD * 10000000 / RATE;
   
   mean = var = 0.0;

   m = (Model *) calloc (1, sizeof (Model));
   um.mtail = um.mhead = m;  
   um.totalframe = um.nState = um.nModel = 0;
   
   while (!feof(labfp)) {
      GetToken (labfp,buf);
      if (!isalnum(buf[0])) break;
      if (isdigit(buf[0]))
         hastime = TRUE;
      else 
         hastime = FALSE;
      
      if (hastime) {
         if (gp->algnst) {
            start = atoi(buf);
            GetToken(labfp, buf);
            end = atoi(buf);
            GetToken(labfp, buf);
            GetToken(labfp, buf);
         }
         else if (gp->algnph) {
            start = atoi(buf);
            GetToken(labfp, buf);
            end = atoi(buf);
            GetToken(labfp, buf);
         }
         else {
            do {
               GetToken(labfp, buf);
            } while (isdigit(buf[0]));
         }
      }
      
      m->name = strdup(buf);
      
      if (hastime && gp->algnph) {
         m->durpdf = SearchTree(m->name, ts->thead[DUR]->root);
         FindDurPDF(m, ms, gp->RHO, diffdur);
         nf = 0;
         for (state=2; state<=ms->nstate+1; state++)
            nf += m->dur[state];
           
         fprintf(stderr, ">>>nf=%d %d\n", nf, (end-start)/rate);
         
         f = (float)(end-start)/(rate*nf);
         m->totaldur = 0;
         
         for (state=2; state<=ms->nstate+1; state++) {
            nf = (int)(f*m->dur[state]+0.5);
            if (nf<=0)  nf=1;
            fprintf(stderr, "%d: %d %f %d\n", state, m->dur[state], f, nf);
            m->dur[state] = nf;
            m->totaldur += m->dur[state];
         }
         um.totalframe += m->totaldur;
      }
      else if (hastime && gp->algnst) {
         m->dur = (int *) calloc(ms->nstate, sizeof(int));
         m->dur -= 2;
         m->dur[2] = (end-start)/rate;
         m->totaldur = m->dur[2];
         um.totalframe += m->dur[2];
         
         for (state=3; state<=ms->nstate+1; state++) {
            GetToken(labfp, buf);
            start = atoi(buf);
            GetToken(labfp, buf); 
            end = atoi(buf);
            GetToken(labfp, buf);
            m->dur[state] = (end-start)/rate;
            m->totaldur += m->dur[state];
            um.totalframe  += m->dur[state];
         }
      } 
      else {
         m->durpdf = SearchTree(m->name, ts->thead[DUR]->root);   
         if (gp->LENGTH==0) {
            FindDurPDF(m, ms, gp->RHO, diffdur);
            um.totalframe += m->totaldur;
         }
         else {   /* if total length of generated speech is specified */
            for (state=2; state<=ms->nstate+1; state++) {
               mean += ms->durpdf[m->durpdf][state];
               var  += ms->durpdf[m->durpdf][state+ms->nstate];
            }
         }
      }
      
      /* for excitation */
      m->lf0pdf      = (int *)     calloc (ms->nstate, sizeof(int)     );
      m->lf0mean     = (float **)  calloc (ms->nstate, sizeof(float *));
      m->lf0variance = (float **)  calloc (ms->nstate, sizeof(float *));
      m->voiced      = (Boolean *) calloc (ms->nstate, sizeof(Boolean));
      
      m->lf0pdf -= 2;  m->lf0mean -= 2;  m->lf0variance -= 2;  m->voiced -= 2;
      
      for (tree=ts->thead[LF0],state=2; tree!=ts->ttail[LF0]; tree=tree->next,state++) {
         m->lf0pdf[state] = SearchTree(m->name, tree->root);
         FindLF0PDF(state, m, ms, gp->UV);
      }

      /* for spectrum */
      m->mceppdf      = (int *)    calloc (ms->nstate, sizeof(int)     );
      m->mcepmean     = (float **) calloc (ms->nstate, sizeof(float *));
      m->mcepvariance = (float **) calloc (ms->nstate, sizeof(float *));
      
      m->mceppdf -= 2;  m->mcepmean -= 2;  m->mcepvariance -= 2;
      
      for (tree=ts->thead[MCP],state=2; tree!=ts->ttail[MCP]; tree=tree->next,state++) {
         m->mceppdf[state] = SearchTree(m->name, tree->root);
         FindMcpPDF(state, m, ms);
      }
      
      m->next = (Model *) calloc(1, sizeof(Model));
      m = um.mtail = m->next;
      
      um.nModel++;
      um.nState+=ms->nstate;
   }
   
   if (gp->LENGTH > 0 && gp->LENGTH < um.nState) {
      fprintf(stderr, "Specified length of generated speech is too short ! (this sentence HMM is composed from %d states)\n", um.nState);
      fprintf(stderr, "Please specify more than %.1f seconds.\n", (float)(um.nState*FPERIOD)/RATE);
      exit(1);
   }
   
   /* if total length of generated speech is specified */
   /* compute RHO */
   if (gp->LENGTH>0) {
      gp->RHO = (gp->LENGTH - mean)/var;
      /* compute state duration for each state */
      for (m=um.mhead; m!=um.mtail; m=m->next) {
         FindDurPDF(m, ms, gp->RHO, diffdur);
         um.totalframe += m->totaldur;
      }
   }
   
   /* Output label information */
   OutLabel(&um, gp->XIMERA);
   
   pdf2speech(rawfp, lf0fp, mcepfp, mceppst, lf0pst, gp, ms, &um, vs);
}

/* -------------------- End of "hts_engine.c" -------------------- */
